using System;
using System.Collections;
using System.Data;
using System.Globalization;
using System.IO;
using System.Threading;

namespace Banshee 
{
    public abstract class Watch {
        protected string musicFolder;
        
        protected ArrayList toImport;
        protected ArrayList toRemove;
        
        public Watch(ArrayList im, ArrayList rm, string folder)
        {
            musicFolder = folder;
            
            toImport = im;
            toRemove = rm;
        }
        
        public bool RecurseDirectory(string path)
        {
            DirectoryInfo di;
    		
            try {
                di = new DirectoryInfo(path);
            } catch(Exception) {
                return false;
            }
    		
            if(!di.Exists)
                return false;
    		
    		if(!AddWatch(path))
                return false;
    		
            foreach(DirectoryInfo sdi in di.GetDirectories()) {
                if(!sdi.Name.StartsWith(".")) 
                    RecurseDirectory(path + "/" + sdi.Name);
            }
            return true;
        }
                                   
        protected void UniqueAdd(ArrayList aList, string item)
		{
            string cur;
            bool added = false;
            
            for(int i = 0; i < aList.Count; i++) {
                cur = aList[i] as string;
            
                if(IsSubfolder(item, cur))
                    return;
                
                if(IsSubfolder(cur, item)) {
                    aList[i] = item;
                    added = true;
                }
            }
            
            if(!added)
                aList.Add(item);
		}
		
		protected bool IsSubfolder(string f1, string f2)
		{
            CompareInfo comp =  CultureInfo.InvariantCulture.CompareInfo;

            return comp.IsPrefix(f1, f2);
		}
		
		public abstract bool IsWatching(string path);
        
        public abstract bool AddWatch(string path);
        
        public abstract void Stop();
    }
    
    public sealed class InotifiedWatch : Watch {
    
        private bool verbose = false;
                 
        private bool HasFlag(Inotify.EventType type, Inotify.EventType val)
        {
	   	   return ((type & val) == val);
        }
    
        public InotifiedWatch(ArrayList im, ArrayList rm, string folder) : base(im, rm, folder) 
        {
            RecurseDirectory(StringUtil.UriToFileName(musicFolder));
        
            Inotify.Verbose = verbose;
            Inotify.Start();
        }
        
        public override bool IsWatching(string path)
        {
            return Inotify.IsWatching(path);
        }
       
        public override bool AddWatch(string path)
        {            
           if(IsWatching(path) || !Directory.Exists(path))
                return false;
                
            if(verbose)
        	   Console.WriteLine ("Adding watch to {0}", path);
            
       		Inotify.Subscribe (path, OnInotifyEvent, Inotify.EventType.CloseWrite | 
   		          Inotify.EventType.MovedFrom | Inotify.EventType.MovedTo |
   		          Inotify.EventType.Create | Inotify.EventType.Delete);
   		    
   		    return true;
        }
        
        public override void Stop()
		{
            Inotify.Stop();
		}
        
        private void OnInotifyEvent(Inotify.Watch watch, string path, string subitem, 
                                            string srcpath, Inotify.EventType type)
        {
            if(verbose)
    			Console.WriteLine("Got event ({03}) {0}: {1}/{2}", type, path, subitem,
	       						   srcpath);

			string fullPath = Path.Combine(path, subitem);

			lock(this) {
			
				if(HasFlag(type, Inotify.EventType.MovedTo) ||
				    HasFlag(type, Inotify.EventType.CloseWrite))
				{
					UniqueAdd(toImport, fullPath);
						
					if(srcpath != null)
						UniqueAdd(toRemove, srcpath);
						
				} else if (HasFlag (type, Inotify.EventType.Create))
						 /*HasFlag (type, Inotify.EventType.IsDirectory)) */
			    {
					UniqueAdd(toImport, fullPath);
				} else if (HasFlag (type, Inotify.EventType.Delete) ||
						 HasFlag (type, Inotify.EventType.MovedFrom))
				{
					UniqueAdd(toRemove, fullPath);
				}
			}
		}	
 	}
 	
 	public sealed class FSWWatch : Watch
    {
        private Hashtable watchMap;
        private bool verbose = false;
        
        public FSWWatch(ArrayList im, ArrayList rm, string folder) : base(im, rm, folder)
        {
            watchMap = new Hashtable();
            RecurseDirectory(StringUtil.UriToFileName(musicFolder));
        }
        
                
        public override bool IsWatching(string path)
        {
            return watchMap.ContainsKey(path);
        }
        
        public override bool AddWatch(string path)
        {
           if(IsWatching(path) || !Directory.Exists(path))
                return false;
                
            if(verbose)
        	   Console.WriteLine ("Adding watch to {0}", path);
            
            FileSystemWatcher watcher = new FileSystemWatcher();
            
            watcher.Path = path;
            watcher.NotifyFilter = NotifyFilters.LastWrite 
                | NotifyFilters.FileName | NotifyFilters.DirectoryName;
            watcher.IncludeSubdirectories = false;
            watcher.Filter = "";

            watcher.Changed += new FileSystemEventHandler(OnChanged);
            watcher.Created += new FileSystemEventHandler(OnChanged);
            watcher.Deleted += new FileSystemEventHandler(OnDeleted);
            watcher.Renamed += new RenamedEventHandler(OnRenamed);

            watcher.EnableRaisingEvents = true;
            
            watchMap.Add(path, watcher);
            
            return true;
        }
        
        public override void Stop()
        {
        }

 
        private void OnChanged(object source, FileSystemEventArgs e)
        {
            lock(this) {
                UniqueAdd(toImport, e.FullPath);
            }
        }
        
        private void OnDeleted(object source, FileSystemEventArgs e)
        {
            lock(this) {
                UniqueAdd(toRemove, e.FullPath);
            }
        }

        private void OnRenamed(object source, RenamedEventArgs e)
        {
            lock(this) {
                UniqueAdd(toImport, e.FullPath);
                UniqueAdd(toRemove, e.OldFullPath);
            }
        }
    }
    
    public class Watcher {
        private ArrayList toImport;
        private ArrayList toRemove;
        
        private Thread updateThread;
        
        private Watch watch;
        
        public Watcher()
        {
            toImport = new ArrayList();
            toRemove = new ArrayList();
        
            updateThread = new Thread(new ThreadStart(Update));
        
            if(Inotify.Enabled) {
                Console.WriteLine("The power of inotify!");
                watch = new InotifiedWatch(toImport, toRemove, 
                            Core.GconfClient.Get(GConfKeys.LibraryLocation) as string);
            } else {
                watch = new FSWWatch(toImport, toRemove, 
                            Core.GconfClient.Get(GConfKeys.LibraryLocation) as string);
            }
            
            updateThread.Start();
                
        }
        
        private void Update()
        {
            while(true) {
    		    lock(watch) {
                    if(toRemove.Count != 0) {
                        Console.WriteLine("toRemove begin");
                        
                        string query = " FROM TRACKS WHERE";                     
    		    
            		    foreach(string s in toRemove) {
            		        Console.WriteLine(s as string);
                            query +=" Uri LIKE \"file://" + s + "/%\"";
                            query += " OR Uri LIKE \"file://" + s + "\"";
                            
                            query += " OR";                            
                        }
                        
                        Console.WriteLine("toRemove end");
                        
                        query = query.Substring(0, query.Length - 3);
                        
                        string selectQuery = "SELECT TrackID, Uri" + query;
                        string deleteQuery = "DELETE" + query;
                                               
                        IDataReader reader = Core.Library.Db.Query(selectQuery);
                        while(reader.Read()) {
                            Core.Library.Remove(Convert.ToInt32(reader[0] as string), 
                                                new System.Uri(reader[1] as string));
                        }                       
                        Core.Library.Db.Execute(deleteQuery);
                      
                        toRemove.Clear();
                    }
                    
                    if(toImport.Count != 0) {
                        FileLoadTransaction transaction = 
                            new FileLoadTransaction(null, true, true);
                            
                        Console.WriteLine("toImport begin");

              		    foreach(string s in toImport) {
                            Console.WriteLine(s);
                            transaction.AddPath(s);
                            watch.RecurseDirectory(s);
                        }
                        
                        Console.WriteLine("toImport end");
                        
                        transaction.Register();
                        
                        toImport.Clear();
                    }
                }
                Thread.Sleep(5000);
		    }
		}
		
		public void Stop()
		{
            updateThread.Abort();
            watch.Stop();
		}
    }
}
